from shapely.geometry import LinearRing
from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPoint
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon

GeomTypeSet = (
  Point, MultiPoint,
  Polygon, MultiPolygon,
  LineString, MultiLineString,
  LinearRing
)



from .util import get_shortest_element


def get_single_list(start, length, step):
  """
  获取单个起点的位置端点列表，如果起始位置不是零，则会返回0到起始位置的列表，如：

  >>> get_single_list(0, 7, 3) == [[0, 3], [3, 6], [6, 7]]
  >>> get_single_list(1, 7, 3) == [[0, 1], [1, 4], [4, 7]]

  :param start: 起始位置
  :param length: 总长度
  :param step: 步长
  :return:
  """
  if start > 0:
    res = [[0, start]]
  else:
    res = []
  for i in range(start, length, step):
    if i + step <= length:
      res.append([i, i + step])
    else:
      res.append([i, length])
  return res


def get_breaks(length, step):
  """
  通过长度和步长获所有可能的取端点列表的列表，如：

  >>> get_breaks(10, 3) == [
  >>>   [[0, 3], [3, 6], [6, 9], [9, 10]],
  >>>   [[0, 1], [1, 4], [4, 7], [7, 10]],
  >>>   [[0, 2], [2, 5], [5, 8], [8, 10]]
  >>> ]
  """
  res = []
  for start in range(0, step):
    res.append(get_single_list(start, length, step))
  return res


def get_list_by_position(string: str, breaks: list):
  """
  按照位置信息将字符串拆解为多个字符串的列表

  :param string: 需要拆分的字符串
  :param breaks: 位置集合列表的列表（左闭右开），[[1, 5], [5, 10], [10, 15]]
  :return:
  """
  res = []
  for brk in breaks:
    res.append(string[brk[0]:brk[1]])
  return res


def drop_repeat_element(strs: list):
  """删除列表中连续重复的元素"""
  s = ''
  last = ''
  for i in strs:
    if i != last:
      s += i
      last = i
  return s


def drop_repeat_string_by_step(string, step):
  """根据指定的小块长度，获取去重后的字符串"""
  _list = []
  length = len(string)
  for breaks in get_breaks(length, step):
    res = get_list_by_position(string, breaks)
    str_drop = drop_repeat_element(res)
    _list.append(str_drop)

  return get_shortest_element(_list)


def drop_repeat_string(string,
                       min_length=3,
                       max_length=None):
  """
  删除连续重复的字符串

  按照step从大到小删除重复字符，返回去重后最短的字符串"""
  if len(string) < min_length * 2:
    return string
  _list = []
  if not max_length:
    max_length = len(string) // 2
  for step in range(max_length, min_length - 1, -1):
    str_drop = drop_repeat_string_by_step(string, step)
    _list.append(str_drop)
    string = get_shortest_element(_list)
  return string



import random
import re

import numpy as np
import pandas as pd

from ..resource.patterns import Pattern
from .util import all_year_old
from .util import random_date


def get_check_code(id_number) -> str:
  mapping = {
    0: '1', 1: '0', 2: 'X', 3: '9', 4: '8',
    5: '7', 6: '6', 7: '5', 8: '4', 9: '3', 10: '2',
  }
  weight = [7, 9, 10, 5, 8, 4, 2, 1, 6, 3, 7, 9, 10, 5, 8, 4, 2]
  ils = [int(i) for i in id_number[:17]]
  return mapping.get(sum(np.multiply(weight, ils)) % 11)


class IDNumber:
  def __init__(self, id_number: str):
    self.id_number = self.format_id_number(id_number)
    self.__birthday = self.id_number[6:14]
    self.__gender = self.id_number[-2]

  @property
  def value(self):
    return self.id_number

  @staticmethod
  def is_valid(id_number, check_code=False) -> bool:
    """判断身份证号是否有效"""
    string = str(id_number)
    if re.match(Pattern.ID_number, string):
      if check_code:
        if get_check_code(string) != string[-1]:
          return False
      return True
    return False

  @staticmethod
  def not_valid(id_number, check_code=False) -> bool:
    """判断身份证号是否无效"""
    return not IDNumber.is_valid(id_number, check_code == check_code)

  def format_id_number(self, id_number) -> str:
    """对身份证号进行校验并转为字符串格式"""
    if isinstance(id_number, (float, int)):
      id_number = str(int(id_number))
    if self.is_valid(id_number):
      return id_number
    raise ValueError('身份证号格式错误')

  @property
  def check_code(self):
    return get_check_code(self.id_number)

  @property
  def birthdate(self):
    """出生日期"""
    return pd.to_datetime(self.__birthday, format='%Y%m%d')

  def birthday(self, format: str = '%Y-%m-%d') -> str:
    """出生日期"""
    return self.birthdate.strftime(format=format)

  @property
  def birth_int(self) -> int:
    """出生日期"""
    return int(self.birthdate.strftime(format='%Y%m%d'))

  @property
  def age(self) -> int:
    """年龄"""
    return all_year_old(self.birthdate)

  @property
  def gender(self) -> str:
    """性别"""
    return '女' if int(self.__gender) % 2 == 0 else '男'

  @property
  def sex(self) -> str:
    return self.gender

  @classmethod
  def random(cls):
    from ..resource.area_code import id_number_sbfix
    _6 = random.choice(list(id_number_sbfix.keys()))
    _birthday = random_date()
    _ord = random.randint(10, 99)
    _gender = random.randint(0, 9)
    id_number = f'{_6}{_birthday}{_ord}{_gender}'
    check_code = get_check_code(id_number)
    return id_number + check_code



import re
import warnings

from ..geometry.df import mark_tags_v2
from .strings import drop_repeat_string
from .util import union_str


def make_building_address(
    df,
    *,
    c_hao,
    c_lu=None,
    c_nong=None,
    c_xiaoqu=None,
    c_dst='楼宇地址'):
  """
  根据【路弄号】或【小区地址和号】生成新版楼宇地址，格式为xx路xx弄xx号

  Args:
    df: dataframe
    c_hao: 号/楼栋号的列名
    c_lu: 路的列名
    c_nong: 弄的列名
    c_xiaoqu: 小区的列名，合适为"xx路xx弄[xx小区]"，用于提取路和弄
    c_dst: 生成楼宇地址的列名，默认为"楼宇地址"
  """
  for c in [c_hao, c_lu, c_xiaoqu]:
    if c and any(df[c].isna()):
      raise ValueError(f'{c}不可为空')
  if c_dst in df:
    warnings.warn(f'原数据中已存在{c_dst}列，已替换')
  if all([c_lu, c_nong]):
    df[c_dst] = df.apply(
        lambda r: union_str([r[c_lu], r[c_nong], r[c_hao]]), axis=1)
  elif c_xiaoqu:
    df[c_dst] = df.apply(
        lambda r: drop_repeat_string(
            union_str(
                [r[c_xiaoqu].split('[')[0], r[c_hao]]
            )
        ),
        axis=1
    )
  else:
    raise KeyError('[lu, nong] 和 [xiaoqu]不能同时缺失')
  return df


def filter_dup_building(df,
                        xiaoqu,
                        c_building='楼宇地址',
                        c_xiaoqu='小区名称'):
  """
  通过小区边界和楼宇边界的空间关联，筛选同一小区的相同楼栋号
  Args:
    df: 楼宇边界
    xiaoqu: 小区边界
    c_building: 楼宇地址列名
    c_xiaoqu: 小区名称列名
  """
  if not xiaoqu[c_xiaoqu].is_unique:
    raise ValueError(f'{c_xiaoqu}存在重复值')

  df = mark_tags_v2(df, xiaoqu, col_list=[c_xiaoqu])

  df = df[df[c_xiaoqu].notna() & df[c_building].notna()]

  return df[df.duplicated(subset=[c_xiaoqu, c_building])]


class BuildingAddress:
  def __init__(self, building: str):
    self.building = building

  def is_valid(self, building):
    if re.match('^.*路(\d+弄)?\d+-?\d*号楼?$', building):
      return True
    return False

  @property
  def road(self):
    """路"""
    if road_list := re.findall('^.+路', self.building):
      return road_list[0]

  @property
  def lane(self):
    """弄"""
    if lane_list := re.findall('路(.+弄)', self.building):
      return lane_list[0]

  @property
  def num(self):
    """号"""
    if num_list := re.findall('(?:.+路.+弄|.+路)(.+号楼?$)', self.building):
      return num_list[0]
    if num_list := re.findall('\d+号楼?', self.building):
      return num_list[0]

  @property
  def lu(self):
    """路"""
    return self.road

  @property
  def long(self):
    """弄"""
    return self.lane

  @property
  def nong(self):
    """弄"""
    return self.lane

  @property
  def hao(self):
    """号"""
    return self.num

  @property
  def std(self):
    return union_str([self.road, self.lane, self.num])



import logging
import uuid
import warnings

import pandas as pd
from requests.exceptions import ConnectionError
from requests.exceptions import ConnectTimeout
from tqdm import tqdm

from ..util.util import is_empty
from .amap import get_address_amap
from .amap import get_place_amap
from .baidu import get_address_baidu
from .baidu import get_place_baidu


def geocode(*,
            address,
            city,
            source,
            srs='wgs84',
            key_baidu=None,
            key_amap=None):
  """
  Args:
      address: 地址或项目名称关键词
      city: 城市
      source: 接口选择。baidu/amap：百度或高德的地理编码接口，baidu_poi/amap_poi：百度或高德的地点检索接口
      srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
      key_baidu: 百度接口的key，公共key失效后可自行传入
      key_amap: 高德接口的key，公共key失效后可自行传入
  """
  if source == 'baidu':
    return get_address_baidu(city=city, address=address, srs=srs, key=key_baidu)
  elif source == 'baidu_poi':
    return get_place_baidu(city=city, keywords=address, srs=srs, key=key_baidu)
  elif source == 'amap':
    return get_address_amap(city=city, address=address, srs=srs, key=key_amap)
  elif source == 'amap_poi':
    return get_place_amap(city=city, keywords=address, srs=srs, key=key_amap)
  else:
    return ValueError('source参数错误')


def geocode_best_poi(city,
                     keywords,
                     sig=80,
                     address_geocode=False,
                     srs='wgs84',
                     key_baidu=None,
                     key_amap=None):
  """
  获取最优的地点检索结果
  Args:
      city: 城市
      keywords: POI名称
      sig: 解析结果相似度，默认为80，范围0-100
      address_geocode: 是否使用地理编码补全，默认False
      srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
      key_baidu: 百度接口的key，公共key失效后可自行传入
      key_amap: 高德接口的key，公共key失效后可自行传入
  """
  if is_empty(keywords):
    return None, None, None, None, None
  res = {
    'lng': None,
    'lat': None,
    'rv': None,
    'score': None,
  }
  source = None
  try:
    ar = get_place_amap(city, keywords, srs=srs, key=key_amap)
    if ar['score'] >= 90:
      res, source = ar, 'amap_poi'
    else:
      br = get_place_baidu(city, keywords, srs=srs, key=key_baidu)
      if (br['score'] >= ar['score']) and (br['score'] >= sig):
        res, source = br, 'baidu_poi'
      elif (ar['score'] >= br['score']) and (ar['score'] >= sig):
        res, source = ar, 'amap_poi'
      elif address_geocode:
        br = get_address_baidu(city, keywords, srs=srs, key=key_baidu)
        if br['score'] >= 90:
          res, source = br, 'baidu_geocode'
        else:
          ar = get_address_amap(city, keywords, srs=srs, key=key_amap)
          res, source = ar, 'amap_geocode'
    return res['lng'], res['lat'], res['rv'], res['score'], source

  except (ConnectionError, ConnectTimeout, TimeoutError) as e:
    warnings.warn(f'接口超时，【{city}】：【{keywords}】,{e}')
    return None, None, None, None, None


def geocode_best_address(city, address, srs='wgs84', key_baidu=None,
                         key_amap=None):
  """获取最优的地理编码结果"""
  if is_empty(address):
    return None, None, None, None, None
  try:
    br = get_address_baidu(city, address, srs=srs, key=key_baidu)
    if br['score'] >= 90:
      res, source = br, 'baidu_geocode'
    else:
      ar = get_address_amap(city, address, srs=srs, key=key_amap)
      res, source = ar, 'amap_geocode'
    return res['lng'], res['lat'], res['rv'], res['score'], source

  except (ConnectionError, ConnectTimeout, TimeoutError) as e:
    warnings.warn(f'接口超时，【{city}】：【{address}】,{e}')
    return None, None, None, None, None


def geocode_df(df: pd.DataFrame,
               by: str,
               city: (str, list),
               address_type: str,
               sig: int = 80,
               address_geocode: bool = False,
               with_detail: bool = True,
               srs: str = 'wgs84',
               key_baidu=None,
               key_amap=None):
  """
  对dataframe进行geocoding
  Args:
      df:
      by: 项目名称所在的列
      city: 城市，list表示城市列如['城市']， str表示全部使用该城市如'上海'
      address_type: 'poi'或'address'，不同的地理类型对应不同的接口，精确度不同
      sig: 解析结果相似度，默认为80，范围为0-100，仅当address_type为poi时生效
      address_geocode: poi解析时是否使用地理编码补全，默认False，仅当address_type为poi时生效
      with_detail: 是否返回详情信息，包括返回值、得分、接口来源等
      srs: 返回的坐标系，可选wgs84, bd09, gcj02，默认wgs84
      key_baidu: 百度接口的key，公共key失效后可自行传入
      key_amap: 高德接口的key，公共key失效后可自行传入
  """
  temp_city_col = uuid.uuid1()

  if isinstance(city, str):
    df[temp_city_col] = city
  elif isinstance(city, list):
    df[temp_city_col] = df[city]
  else:
    raise TypeError(f'city类型错误，请传入list或str类型')

  on = [temp_city_col, by]
  df_temp = df[on].drop_duplicates().reset_index(drop=True)

  for i in tqdm(df_temp.index):
    try:
      kw = df_temp[by][i]
      ct = df_temp[temp_city_col][i]
      if address_type == 'poi':
        rv = geocode_best_poi(
            ct, kw,
            sig=sig,
            address_geocode=address_geocode,
            srs=srs,
            key_baidu=key_baidu,
            key_amap=key_amap)
      elif address_type == 'address':
        rv = geocode_best_address(ct, kw, srs=srs,
                                  key_baidu=key_baidu, key_amap=key_amap)
      else:
        raise ValueError('参数address_type错误，可选参数为"poi"或"address"')
      df_temp.loc[
        (df_temp[temp_city_col] == ct) & (df_temp[by] == kw),
        ['lng', 'lat', 'rv', 'geocode_score', 'geocode_type']] = rv
    except Exception as e:
      logging.warning(f'{e},{kw}')

  df = df.merge(df_temp, on=on, how='left')
  if not with_detail:
    del df['rv'], df['geocode_score'], df['geocode_type']
  del df[temp_city_col]
  return df



import datetime
import logging
import os
import zipfile

from .util import ensure_list


def ext(filepath):
  """扩展名"""
  return os.path.splitext(filepath)[1]


def fn(filepath):
  """路径及文件名（不含扩展名）"""
  return os.path.splitext(filepath)[0]


def split_path(filepath, abs=False):
  """将文件路径拆分为文件夹路径、文件名、扩展名三部分"""
  if abs:
    filepath = os.path.abspath(filepath)
  basename = os.path.basename(filepath)
  dir_path = os.path.dirname(filepath)
  return dir_path, fn(basename), ext(basename)


def remove_dir(filepath):
  """
  删除某一目录下的所有文件或文件夹
  :param filepath: 路径
  :return:
  """
  import shutil
  del_list = os.listdir(filepath)
  for f in del_list:
    file_path = os.path.join(filepath, f)
    if os.path.isfile(file_path):
      os.remove(file_path)
    elif os.path.isdir(file_path):
      shutil.rmtree(file_path)
  shutil.rmtree(filepath)


def dir2zip(filepath, delete_exist=False, delete_origin=False):
  """压缩文件夹"""
  zfn = filepath + '.zip'
  if delete_exist:
    if os.path.exists(zfn):
      os.remove(zfn)
      print(f'文件已存在，delete {zfn}')
  print(f'saving {zfn}')
  z = zipfile.ZipFile(zfn, 'w', zipfile.ZIP_DEFLATED)
  for dirpath, dirnames, filenames in os.walk(filepath):
    for filename in filenames:
      filepath_out = os.path.join(dirpath, filename)
      filepath_in = os.path.join(os.path.split(dirpath)[-1], filename)
      z.write(filepath_out, arcname=filepath_in)
  z.close()
  if delete_origin:
    print(f'delete {filepath}')
    remove_dir(filepath)


def get_file_counts(dir_path):
  """获取文件夹中文件的数量"""
  num = 0
  for dirpath, dirnames, filenames in os.walk(dir_path):
    for filename in filenames:
      if filename != '.DS_Store':
        num += 1
  return num


def rm_scratch_file(dir_path, days, rm_hidden_file=False):
  """删除指定天数前修改过的文件"""
  dirs = [
    '/bin', '/sbin', '/usr', '/etc', '/dev', '/Applications', '/Library',
    '/Network', '/System', '/Volumes', '/cores', '/private'
  ]
  for pre in dirs:
    if os.path.abspath(dir_path).startswith(fr'{pre}'):
      raise ValueError('系统目录下的文件不可删除')
  if dir_path == '/':
    raise ValueError('不可指定根目录')
  timeline = datetime.datetime.now() - datetime.timedelta(days)
  # 删除过期文件
  for dirpath, dirnames, filenames in os.walk(dir_path):
    for filename in filenames:
      if not rm_hidden_file:
        continue
      full_path = os.path.join(dirpath, filename)
      m_time = os.path.getmtime(full_path)
      m_time = datetime.datetime.fromtimestamp(m_time)
      if m_time <= timeline:
        os.remove(full_path)
        logging.warning(f'已删除文件：{full_path}')
  # 删除空文件夹
  for dirpath, dirnames, filenames in os.walk(dir_path):
    for dirname in dirnames:
      dir_full_path = os.path.join(dirpath, dirname)
      num = get_file_counts(dir_full_path)
      if num == 0:
        remove_dir(dir_full_path)
        logging.warning(f'已删除空文件夹：{dir_full_path}')


def dir_iter(root, exts: (list, str) = None, abspath=False):
  """
  文件夹中的文件路径生成器，用于遍历文件夹中的文件
  Args:
    root: 文件目录
    exts: 文件扩展名，不指定则返回所有文件
    abspath: 是否返回绝对路径
  """

  for filename in os.listdir(root):
    filepath = os.path.join(root, filename)
    if abspath:
      filepath = os.path.abspath(filepath)
    if exts:
      exts = ensure_list(exts)
      if ext(filepath) in exts:
        yield filepath
    else:
      yield filepath



from .df import auto2shapely
from .df import buffer
from .df import geojson2shapely
from .df import get_area
from .df import lnglat2shapely
from .df import lnglat2wkb
from .df import lnglat2wkt
from .df import mark_tags_v2
from .df import nearest_neighbor
from .df import projection
from .df import shapely2geojson
from .df import shapely2lnglat
from .df import shapely2wkb
from .df import shapely2wkt
from .df import shapely2x
from .df import spatial_agg
from .df import wkb2lnglat
from .df import wkb2shapely
from .df import wkb2wkt
from .df import wkt2lnglat
from .df import wkt2shapely
from .df import wkt2wkb
from .util import distance
from .util import geojson_dumps
from .util import geojson_loads
from .util import get_epsg
from .util import get_epsg_by_lng
from .util import get_inner_point
from .util import infer_geom_format
from .util import is_geojson
from .util import is_shapely
from .util import is_wkb
from .util import is_wkt
from .util import projection_lnglat
from .util import wkb_dumps
from .util import wkb_loads
from .util import wkt_dumps
from .util import wkt_loads



import warnings
from datetime import datetime

import numpy as np
import pandas as pd

from ..util.assertion import assert_series_unique
from ..util.decorator import progress
from ..util.decorator import timer
from ..util.util import and_
from ..util.util import ensure_list
from ..util.util import fuzz_match
from ..util.util import is_empty
from ..util.util import list2dict
from ..util.util import to_float


def best_unique(df: pd.DataFrame,
                key_cols: (list, str),
                value_cols: (str, list) = None,
                filter=False,
                drop_if_null=None) -> pd.DataFrame:
  """
  优化的去重函数：
    为保证数据的完整性，去重时优先去除指定列中的空值

  :param df:
  :param key_cols: 按照哪些列去重
  :param value_cols: 优先去除那些列的空值，该列表是有顺序的
  :param filter:
  :param drop_if_null: 如何处理value_cols内值为空的列；'all'：都为空时删除该列，'any'：任意一列为空时就删除，None：保留空白
  :return:
  """
  key_cols = ensure_list(key_cols)
  if not value_cols:
    value_cols = [i for i in df.columns if i not in key_cols]
  value_cols = ensure_list(value_cols)
  if drop_if_null:
    df = df.dropna(
        subset=value_cols,
        how=drop_if_null
    ).dropna(
        subset=key_cols,
        how='all')
  df = df.sort_values(value_cols, na_position='first')
  df = df.drop_duplicates(key_cols, keep='last', ignore_index=True)
  if filter:
    df = df[key_cols + value_cols]
  return df


def keep_best_unique(df: pd.DataFrame,
                     subset: (list, str),
                     value_cols: (str, list) = None
                     ) -> pd.DataFrame:
  """
  优化的去重函数，为保证数据的完整性，去重时优先去除指定列中的空值
  Args:
    df: 要去重的Dataframe
    subset: 按照哪些列去重
    value_cols: 优先去除那些列的空值，该列表是有顺序的
  """
  subset = ensure_list(subset)

  if not value_cols:
    value_cols = [i for i in df if i not in subset]
  value_cols = ensure_list(value_cols)

  return df.sort_values(
      value_cols, na_position='first'
  ).drop_duplicates(
      subset, keep='last'
  ).sort_index()


def table2dict(df: pd.DataFrame,
               key_col: str = None,
               value_col: (str, list) = None,
               orient: str = 'dict') -> dict:
  """
  DataFrame转字典
  Args:
    df:
    key_col: 生成key的列
    value_col: 生成value的列
    orient: 生成dict的方式，默认'dict',还有 ‘list’, ‘series’, ‘split’, ‘records’, ‘index’
  """
  if not all([key_col, value_col]):
    columns = df.columns.tolist()
    key_col = columns[0]
    value_col = columns[1]
  df = df[df[key_col].notna()].set_index(key_col)
  if isinstance(value_col, list):
    df = df[value_col]
    return df.to_dict(orient=orient)
  else:
    df = df[[value_col]]
    return df.to_dict(orient=orient)[value_col]


def round_by_columns(df, columns: list):
  """对整列进行四舍五入，默认绝对值大于1的数值保留两位小数，小于1 的保留4位"""

  def _round(x):
    if is_empty(x):
      return np.nan
    if abs(x) >= 1:
      return round(x, 2)
    return round(x, 4)

  columns = ensure_list(columns)
  for c in columns:
    df[c] = df[c].apply(lambda x: _round(x))
  return df


def standard(series: (pd.Series, list),
             q: float = 0.01,
             min_score: float = 0,
             minus: bool = False) -> (pd.Series, list):
  if minus:
    series = 1 / (series + 1)
  max_ = series.quantile(1 - q)
  min_ = series.quantile(q)
  series = series.apply(
      lambda x: (x - min_) / (max_ - min_) * (100 - min_score) + min_score)
  series[series >= 100] = 100
  series[series <= min_score] = min_score
  return series


def update_df(df: pd.DataFrame,
              new_df: pd.DataFrame,
              on: (str, list) = None,
              overwrite: bool = True,
              errors: str = 'ignore') -> pd.DataFrame:
  """
  根据某一列更新dataframe里的数据
  Args:
    df: 待升级的数据集
    new_df: 用于更新的DataFrame
    on: （可选参数）用于判断更新哪些行的列
    overwrite : （可选参数）控制如何处理原DataFrame在重叠位置上 **非空** 的值，默认为True

      * True: 默认值；使用 `other` DataFrame中的值覆盖原DataFrame中相应位置的值.
      * False: 只更新原DataFrame中重叠位置数据为 *空* 的值.
    errors: （可选参数）控制如何处理两个DataFrame同一位置都有值的行为，默认为'ignore'

      * 'ignore': 默认值；DataFrame类型 df和other在同一个cell位置都是非NA值，
        使用other中的值替换df中的值。
      * 'raise': target和other都在同一位置包含非NA数据将抛出ValueError异常（'Data overlaps'）。
  """
  if on:
    df.set_index(on, inplace=True)
    df.update(new_df.set_index(on), overwrite=overwrite, errors=errors)
    df.reset_index(inplace=True)
  else:
    df.update(new_df, overwrite=overwrite, errors=errors)
  return df


def date_to(series: pd.Series, mode: str = 'first') -> pd.Series:
  """
  将日期转为当月的第一天或最后一天

  :param series: pd.Series
  :param mode: 'first' or 'last'
  :return:
  """
  from pandas.tseries.offsets import MonthEnd

  def trans(x):
    if x is not pd.NaT:
      y = int(x.year)
      m = int(x.month)
      d = int(x.day)
      return datetime(y, m, d)
    else:
      return None

  series = pd.to_datetime(series)
  if mode == 'first':
    series = series.apply(lambda x: x.replace(day=1))
  elif mode == 'last':
    series = pd.to_datetime(series, format="%Y%m") + MonthEnd(1)
  else:
    raise ValueError(f"{mode}不是正确的参数，请使用 'first' or 'last'")
  return series.apply(trans)


@progress
def fuzz_df(df: pd.DataFrame,
            col: str,
            target_series: (list, pd.Series),
            c_dst=None) -> pd.DataFrame:
  """
  模糊匹配。为DataFrame中的某一列从某个集合中模糊匹配匹配相似度最高的元素
  Args:
    df: 输入的dataframe
    col: 要匹配的列
    target_series: 从何处匹配， list/pd.Series
    c_dst: 关联后输出的列名，默认为原列名+"_target"后缀
  """
  if not c_dst:
    c_dst = f'{col}_target'

  df_temp = df[[col]].drop_duplicates(ignore_index=True)

  df_temp[[
    c_dst, 'normal_score', 'partial_score'
  ]] = df_temp.progress_apply(
      lambda r: fuzz_match(r[col], target_series),
      result_type='expand',
      axis=1)
  return df.merge(df_temp, on=col, how='left')


def series_to_float(series: pd.Series, rex_method: str = 'mean') -> pd.Series:
  """
  pandas.Series: str --> float

  :param series: 要转换的pandas列
  :param rex_method: 计算mean,max,min， 默认为mean
  """
  return series.apply(lambda x: to_float(x, rex_method=rex_method))


def filter_by_df(df: pd.DataFrame, sizer: pd.DataFrame) -> pd.DataFrame:
  """根据一个dataframe筛选另一个dataframe"""
  sizer = sizer.drop_duplicates()
  cond = False
  for i in sizer.index:
    _cond = True
    for c in sizer:
      value = sizer[c][i]
      if pd.isna(value):
        _cond = _cond & df[c].isna()
      else:
        _cond = _cond & (df[c] == value)
    cond = cond | _cond
  return df[cond]


@timer
@progress
def expand_dict(df, c_src):
  """展开字典为多列"""
  return pd.concat(
      [
        df.drop(c_src, axis=1),
        df[c_src].progress_apply(
            lambda x: pd.Series(x, dtype='object')
        ).set_index(df.index)
      ],
      axis=1
  )


@progress
@timer
def split_list_to_row(df: pd.DataFrame, column):
  """将列表列中列表的元素拆成多行"""
  if pd.__version__ >= '1.3.0':
    return df.explode(column, ignore_index=True)
  df[column] = df[column].progress_apply(list2dict)
  return df.drop(columns=column).join(
      expand_dict(
          df[[column]], column
      ).stack(
      ).reset_index(
          level=1, drop=True
      ).rename(column)
  )


def dict2df(data: dict, c_key='key', c_value='value', as_index=False):
  """将字典转为dataframe"""
  df = pd.DataFrame()
  df[c_key] = data.keys()
  df[c_value] = data.values()
  if as_index:
    df.set_index(c_key, inplace=True)
  return df


def is_unique(df: pd.DataFrame, key_cols: (str, list) = None):
  """判断是否唯一"""
  warnings.warn(
      '方法即将停用，请使用"ricco.util.util.is_unique_series"方法替代',
      DeprecationWarning
  )
  if not key_cols:
    key_cols = df.columns.to_list()
  key_cols = ensure_list(key_cols)
  return not df.duplicated(subset=key_cols, keep=False).any()


def one_line(df, key_cols, key_value, value_cols):
  """获取一行重置索引后的数据"""
  cond = and_(*[df[c] == key_value[c] for c in key_cols])
  df = df[cond][[*key_cols, *value_cols]]
  return df.reset_index(drop=True)


@timer
def is_changed(df_old: pd.DataFrame,
               df_new: pd.DataFrame,
               key_cols: (str, list) = None,
               value_cols: (str, list) = None,
               c_res: str = 'is_changed') -> pd.DataFrame:
  """
  判断新旧数据集中的每一条数据是否变化
  Args:
    df_old: 原始数据集
    df_new: 修改后的数据集
    key_cols: 关键列，不指定则以索引列为准
    value_cols: 要对比的列，默认出key_cols之外的其他列
    c_res: 对比结果列名，默认为“is_changed”
  """
  # 数据集及参数检验
  key_cols = ensure_list(key_cols)
  assert_series_unique(df_new, key_cols)
  if not value_cols:
    error_msg = 'Each datasets must have same columns.'
    assert set(df_new.columns) == set(df_old.columns), error_msg
    value_cols = [c for c in df_old if c not in key_cols]
  # 对比
  # 默认所有的都是更改的
  df_new = df_new.copy()
  df_new[c_res] = 'Changed'
  df_temp = df_old[[*key_cols, *value_cols]].merge(
      df_new[[*key_cols, *value_cols, c_res]],
      how='left',
      on=key_cols)
  # 合并后为空的为 NotFound
  df_temp[c_res] = df_temp[c_res].fillna('NotFound')

  # 对比每一个值，如果每一列的每一个值都相同，则认为是没有变化
  cond = and_(
      *[
        (df_temp[f'{c}_x'] == df_temp[f'{c}_y']) |
        (df_temp[f'{c}_x'].isna() & df_temp[f'{c}_y'].isna())
        for c in value_cols
      ]
  )
  df_temp.loc[cond, c_res] = 'NotChange'
  df_temp = df_temp[[*key_cols, c_res]]
  return df_old.merge(df_temp, how='left', on=key_cols)


def df_iter(df: pd.DataFrame, *, chunksize: int = None, parts: int = None):
  if not any([chunksize, parts]):
    raise ValueError(f'chunksize 和 parts必须指定一个')
  size = df.shape[0]
  if chunksize:
    parts = int(size / chunksize) + 1
  else:
    chunksize = int(size / parts) + 1
  for i in range(parts):
    low = i * chunksize
    high = (i + 1) * chunksize
    if i == parts - 1:
      yield df.iloc[low:, :]
    else:
      yield df.iloc[low: high, :]



import json
import warnings
from typing import Union
from typing import List

import geojson
import numpy as np
import pandas as pd
from shapely import wkb
from shapely import wkt
from shapely.errors import GeometryTypeError
from shapely.errors import WKBReadingError
from shapely.geometry import GeometryCollection
from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape
from shapely.geos import WKTReadingError

from ..resource.geometry import GeomTypeSet
from ..util.util import ensure_list
from ..util.util import is_empty
from ..util.util import not_empty


class GeomFormat:
  wkb = 'wkb'
  wkt = 'wkt'
  shapely = 'shapely'
  geojson = 'geojson'
  unknown = 'unknown'


def crs_sh2000():
  """测绘院（上海2000）crs信息"""
  from ..resource.crs import CRS_SH2000
  return CRS_SH2000


def get_epsg_by_lng(lng):
  """通过经度获取epsg代码"""
  lng_epsg_mapping = {
    # 中央经线和epsg代码的对应关系
    75: 4534, 78: 4535, 81: 4536, 84: 4537, 87: 4538,
    90: 4539, 93: 4540, 96: 4541, 99: 4542,
    102: 4543, 105: 4544, 108: 4545, 111: 4546,
    114: 4547, 117: 4548, 120: 4549, 123: 4550,
    126: 4551, 129: 4552, 132: 4553, 135: 4554,
  }
  lng = ensure_list(lng)
  lng = [i for i in lng if not_empty(i)]
  lng = np.median(lng)
  key = min(lng_epsg_mapping.keys(), key=lambda x: abs(x - lng))
  return lng_epsg_mapping.get(key)


def get_lng_by_city(city: str):
  from ..resource.epsg_code import CITY_POINT
  if city in CITY_POINT.keys():
    return CITY_POINT[city]['lng']
  else:
    city += '市'
    if city in CITY_POINT.keys():
      return CITY_POINT[city]['lng']
    else:
      warnings.warn(f'请补充{city}的epsg信息，默认返回经度120')
      return 120


def get_epsg(city: str):
  """根据城市查询epsg代码，用于投影"""
  return get_epsg_by_lng(get_lng_by_city(city))


def projection_lnglat(lnglat, crs_from, crs_to):
  from pyproj import Transformer
  transformer = Transformer.from_crs(crs_from, crs_to)
  return transformer.transform(xx=lnglat[1], yy=lnglat[0])


def wkb_loads(x, hex=True):
  warnings.filterwarnings('ignore',
                          'Geometry column does not contain geometry.',
                          UserWarning)

  if is_empty(x):
    return

  try:
    return wkb.loads(x, hex=hex)
  except (AttributeError, WKBReadingError) as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def wkb_dumps(x, hex=True, srid=4326) -> (str, None):
  if is_empty(x):
    return

  try:
    return wkb.dumps(x, hex=hex, srid=srid)
  except AttributeError as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def wkt_loads(x):
  if is_empty(x):
    return
  try:
    return wkt.loads(x)
  except (AttributeError, WKTReadingError, TypeError) as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def wkt_dumps(x) -> (str, None):
  if is_empty(x):
    return
  try:
    return wkt.dumps(x)
  except AttributeError as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def geojson_loads(x):
  """geojson文本形式转为shapely格式"""
  from simplejson.errors import JSONDecodeError

  if is_empty(x):
    return
  try:
    geom = shape(geojson.loads(x))
    if geom.is_empty:
      return
    return geom
  except (JSONDecodeError, AttributeError, GeometryTypeError, TypeError) as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def geojson_dumps(x) -> (str, None):
  """shapely转为geojson文本格式"""
  if is_empty(x):
    return
  try:
    geom = geojson.Feature(geometry=x)
    return json.dumps(geom.geometry)
  except TypeError as e:
    warnings.warn(f'{e}, 【{x}】')
    return


def is_shapely(x, na=False) -> bool:
  """判断是否为shapely格式"""
  from ..resource.geometry import GeomTypeSet
  if pd.isna(x):
    return na
  if type(x) in GeomTypeSet:
    return True
  else:
    return False


def is_wkb(x, na=False) -> bool:
  """判断是否为wkb格式"""
  if not isinstance(x, str):
    return False
  if pd.isna(x):
    return na
  try:
    wkb.loads(x, hex=True)
    return True
  except WKBReadingError:
    return False


def is_wkt(x, na=False) -> bool:
  """判断是否为wkt格式"""
  if not isinstance(x, str):
    return False
  if pd.isna(x):
    return na
  try:
    wkt.loads(x)
    return True
  except WKTReadingError:
    return False


def is_geojson(x, na=False) -> bool:
  """判断是否为geojson格式"""
  from simplejson.errors import JSONDecodeError

  if not isinstance(x, (str, dict)):
    return False
  if pd.isna(x):
    return na
  try:
    if shape(geojson.loads(x)).is_empty:
      return False
    return True
  except (JSONDecodeError, AttributeError, GeometryTypeError, TypeError):
    return False


def infer_geom_format(x):
  """推断geometry格式"""
  if is_shapely(x):
    return GeomFormat.shapely
  if is_wkb(x):
    return GeomFormat.wkb
  if is_wkt(x):
    return GeomFormat.wkt
  if is_geojson(x):
    return GeomFormat.geojson
  return GeomFormat.unknown


def ensure_multi_geom(geom):
  """将LineString和Polygon转为multi格式"""
  if geom.geom_type == 'LineString':
    return MultiLineString([geom])
  if geom.geom_type == 'Polygon':
    return MultiPolygon([geom])
  return geom


def multiline2multipolygon(multiline_shapely):
  """multiline转为multipolygon，直接首尾相连"""
  coords = []
  for line in multiline_shapely:
    lngs = line.xy[0]
    lats = line.xy[1]
    for i in range(len(lngs)):
      lng, lat = lngs[i], lats[i]
      point = Point((lng, lat))
      if len(coords) >= 1:
        if point != coords[-1]:
          coords.append(point)
      else:
        coords.append(point)
  try:
    return MultiPolygon([Polygon(coords)])
  except ValueError as e:
    warnings.warn(f'{e}，{multiline_shapely}')
    return


def get_inner_point(polygon: Polygon, within=True):
  """返回面内的一个点，默认返回中心点，当中心点不在面内则返回面内一个点"""
  if is_empty(polygon):
    return
  point = polygon.centroid
  if not polygon.is_valid:
    polygon = polygon.buffer(0.000001)
  if polygon.contains(point) or not within:
    return point
  return polygon.representative_point()


def ensure_lnglat(lnglat) -> tuple:
  if isinstance(lnglat, (list, tuple)):
    if all([isinstance(i, (float, int)) for i in lnglat]):
      return lnglat
    else:
      raise TypeError('数据类型错误，经度和纬度都应该为数值型')
  if is_shapely(lnglat):
    geom = lnglat
  elif is_wkt(lnglat):
    geom = wkt_loads(lnglat)
  elif is_wkb(lnglat):
    geom = wkb_loads(lnglat)
  elif is_geojson(lnglat):
    geom = geojson_loads(lnglat)
  else:
    raise ValueError('未知的地理类型')
  return (geom.x, geom.y)


def distance(
    p1: (tuple, str),
    p2: (tuple, str),
    city: str = None,
    epsg_from: int = 4326,
    epsg_to: (str, int) = None):
  """
  计算两个点（经度，纬度）之间的距离，单位：米
  Args:
    p1: 点位1，经纬度或geometry
    p2: 点位2，经纬度或geometry
    city: 所在城市，用于投影
    epsg_from: epsg代码
    epsg_to: 投影epsg代码
  """
  p1, p2 = ensure_lnglat(p1), ensure_lnglat(p2)
  if not epsg_to:
    epsg_to = get_epsg(city) if city else get_epsg_by_lng([p1[0], p2[0]])
  p1 = projection_lnglat(p1, epsg_from, epsg_to)
  p2 = projection_lnglat(p2, epsg_from, epsg_to)
  return Point(p1).distance(Point(p2))


def get_geoms(geo: Union[GeomTypeSet]) -> List:
  """
  将geo中包含的LineString、Point、Polygon取出存放到list中
  Args:
    geo: geometry类型的数据

  Returns:
    List：geo中包含的LineString、Point、Polygon
  """
  list_geo = []
  if geo.is_empty:
    list_geo.append(geo)
  elif hasattr(geo, 'geoms'):
    for g in geo.geoms:
      list_geo.extend(get_geoms(g))
  else:
    list_geo.append(geo)
  return list_geo



import warnings

from ..util.coord_trans import gcj2bd
from ..util.coord_trans import gcj2wgs
from ..util.util import is_empty


class MapUrls:
  # 百度地理编码
  baidu = 'http://api.map.baidu.com/geocoding/v3/'
  # 百度地点检索
  baidu_poi = 'http://api.map.baidu.com/place/v2/search'
  # 高德地理编码
  amap = 'https://restapi.amap.com/v3/geocode/geo'
  # 高德地点检索
  amap_poi = 'https://restapi.amap.com/v3/place/text'
  # 脉策geocode服务
  mdt = 'https://geocode.idatatlas.com/geocode'


class MapKeys:
  amap = '7c14855824549a84c543e48990239f3d'
  baidu = '9Fy1lMHbwpr07WVBFPLw9vpntGUSOUMN'


def gcj2xx(lnglat, srs):
  if srs == 'wgs84':
    return gcj2wgs(float(lnglat[1]), float(lnglat[0]))
  elif srs == 'bd09':
    return gcj2bd(float(lnglat[1]), float(lnglat[0]))
  elif srs == 'gcj02':
    return float(lnglat[1]), float(lnglat[0])
  else:
    raise ValueError('参数输入错误，可选参数为bd09,wgs84,gcj02')


def error_baidu(js):
  if js['status'] != 0:
    warnings.warn(
        '接口错误，状态码【%s】，错误原因请查阅：'
        'https://lbsyun.baidu.com/index.php?title=webapi/appendix' % js[
          'status']
    )


def error_amap(js):
  if js['status'] != '1':
    warnings.warn(
        '接口错误，状态码【%s】，错误原因请查阅：'
        'https://lbs.amap.com/api/webservice/guide/tools/info' % js['infocode']
    )


def fix_address(string):
  if is_empty(string):
    return ''
  return str(string)


def rv_score(city, keywords, rv):
  from fuzzywuzzy import fuzz
  return max(
      fuzz.ratio(rv, keywords.lstrip(city)),
      fuzz.partial_ratio(rv, keywords.lstrip(city))
  )



import logging

import geopandas as gpd
import pandas as pd
from pandas.errors import SettingWithCopyWarning
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from shapely.ops import unary_union

from ricco.etl.transformer import update_df
from ricco.util.assertion import assert_not_null
from shapely.validation import make_valid

from ricco.geometry.util import get_geoms

import warnings
warnings.filterwarnings('ignore', category=UserWarning)
warnings.filterwarnings('ignore', category=SettingWithCopyWarning)
warnings.filterwarnings('ignore', category=RuntimeWarning)
warnings.filterwarnings('ignore', category=FutureWarning)

_desc = '''
  拓扑检查
  可检查: 重叠问题, 自相交问题, geometry缺失问题
  可修复: 重叠问题, 自相交问题
'''


def overlap_check(gdf: gpd.GeoDataFrame, geo_col='geometry') -> gpd.GeoDataFrame:
  """
  面数据重叠检查：
  遍历待检查的geoDataFrame, 两两之间检查是否重叠, 并提取重合部分
  """
  overlay_all = []
  for item in range(len(gdf) - 1):
    overlay_one = gpd.overlay(
        gdf.iloc[item: item + 1],
        gdf.iloc[item + 1:],
        how='intersection',
        keep_geom_type=False,
    )
    if not overlay_one.empty:
      overlay_all.append(overlay_one)
  if not overlay_all:
    logging.warning('所有面数据不存在相交情况')
    return gpd.GeoDataFrame()
  return gpd.GeoDataFrame(pd.concat(overlay_all),
                          crs=4326,
                          geometry=geo_col)


def is_contain(r, geo_col='geometry'):
  if r[geo_col].is_empty:
    return False
  diff = r[geo_col].difference(r['geometry_1']
                               .union(r['geometry_2'])
                               .buffer(1e-7, resolution=1))
  if diff.is_empty:
    return True
  else:
    return False


def fix_contains(gdf: gpd.GeoDataFrame,
                 geo_col='geometry',
                 keep_contains=False):
  """
  处理包含关系面数据，只能处理一层包含关系，
  对于多层面数据包含的情况需要多次调用此方法
  Args:
    gdf: 待处理的数据
    geo_col: 数据的geometry列名
    keep_contains: 是否保留被包含的面数据

  Returns:

  """
  df = (
    gdf[[geo_col]]
    .explode(index_parts=True)
    .reset_index(names=['index', 'overlap_index'])
  )
  overlay_all = overlap_check(df, geo_col)
  if overlay_all.empty:
    return 0, gdf

  # 找出被包含的面,会有两种情况：
  # 1.直接被另一个面包含
  # 2.被多个面一起包含
  # 本面与其他面相交情况
  contains_res = (
    overlay_all[['index_1', 'overlap_index_1', 'geometry']]
    .dissolve(by=['index_1', 'overlap_index_1'], as_index=False)
    .rename(columns={'overlap_index_1': 'overlap_index',
                     'index_1': 'index',
                     geo_col: 'geometry_1'})
    .merge(df,
           on=['index', 'overlap_index'],
           how='right')
  )
  # 其他面与本面相交情况
  contains_res = (
    overlay_all
    .dissolve(by=['index_2', 'overlap_index_2'], as_index=False)
    .rename(columns={'overlap_index_2': 'overlap_index',
                     'index_2': 'index',
                     geo_col: 'geometry_2'})
    .merge(contains_res,
           on=['index', 'overlap_index'],
           how='right')
  )
  contains_res = gpd.GeoDataFrame(contains_res,
                                  crs=4326,
                                  geometry=geo_col)
  contains_res['geometry_1'] = contains_res['geometry_1'].fillna(Polygon())
  contains_res['geometry_2'] = contains_res['geometry_2'].fillna(Polygon())
  contains_res['is_contains'] = contains_res.apply(is_contain,
                                                   args=(geo_col,),
                                                   axis=1)
  contains_res = contains_res[['overlap_index', 'index', 'is_contains', geo_col]]
  contains_geo = contains_res[contains_res['is_contains']]
  contains_num = len(contains_geo)
  if not contains_num:
    return 0, gdf
  logging.warning(f'被包含的面有{contains_num}个: %s',
                  contains_geo[['index', geo_col]])
  if keep_contains:
    contains_geo = contains_res[contains_res['is_contains']]
    contains_geo[geo_col] = contains_geo[geo_col].buffer(1e-7)
    fix_contain = contains_res[~contains_res['is_contains']].overlay(
        contains_geo,
        how='difference'
    )
    update_df(contains_res,
              fix_contain,
              on=['index', 'overlap_index'],
              overwrite=True)
  else:
    contains_res.loc[contains_res['is_contains'], geo_col] = Polygon()
  contains_res = contains_res.groupby('index').agg({geo_col: lambda df: unary_union(df.values)})
  res = gdf.drop(geo_col, axis=1).join(contains_res, how='left')
  return contains_num, gpd.GeoDataFrame(res, crs=4326, geometry=geo_col)


def fix_overlap(gdf: gpd.GeoDataFrame,
                geo_col='geometry',
                fill_intersects=False,
                keep_contains=False) -> gpd.GeoDataFrame:
  """
  修复重叠面数据:
  进行重叠检查, 将原始数据和得到重叠部分略微buffer做差集, 以消除重叠部分
  Args:
    gdf: 需要修复的面数据
    geo_col: 面数据geometry列名
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会随机分配到其中一个面内
    keep_contains: fill_intersects为True时才有效，即是否保留被包含的面
  Returns:
    object: 

  """
  if not fill_intersects:
    overlay_all = overlap_check(gdf[[geo_col]], geo_col)
    overlay_all[geo_col] = overlay_all[geo_col].buffer(1e-7, resolution=1)
    res = gpd.overlay(gdf[[geo_col]], overlay_all, how='difference')
    gdf_clear = gdf.drop(geo_col, axis=1).join(res, how='left')
    gdf_clear[geo_col] = gdf_clear[geo_col].fillna(Polygon())
    return gpd.GeoDataFrame(gdf_clear, geometry=geo_col, crs=4326)
  else:
    contains_num, gdf_fix_contain = fix_contains(gdf, geo_col, keep_contains)
    # 可能会出现多层包含关系，多次调用方法处理这种情况
    while contains_num:
      contains_num, gdf_fix_contain = fix_contains(gdf_fix_contain,
                                                   geo_col,
                                                   keep_contains)

    df = (
      gdf_fix_contain[[geo_col]]
      .explode(index_parts=True)
      .reset_index(names=['index', 'overlap_index'])
    )
    df['index'] = df['index'].astype('int')
    df = gpd.GeoDataFrame(df, geometry=geo_col, crs=4326,)
    intersects = overlap_check(df, geo_col)
    if intersects.empty:
      return gdf_fix_contain
    intersects = (
      intersects
      .groupby(['overlap_index_1', 'index_1'])
      .agg({geo_col: lambda df: unary_union(df.values)})
      .reset_index()
      .rename(columns={'overlap_index_1': 'overlap_index',
                       'index_1': 'index',
                       geo_col: 'intersects'})
      .merge(df,
             on=['index', 'overlap_index'],
             how='right')
    )

    intersects[geo_col] = intersects.apply(
        lambda r: r[geo_col].difference(
            r['intersects'].buffer(1e-7,
                                   resolution=1))
        if pd.notna(r['intersects'])
        else r[geo_col],
        axis=1)

    res = intersects.groupby('index').agg({geo_col: lambda df: unary_union(df.values)})
    gdf_clear = gdf.drop(geo_col, axis=1).join(res, how='left')
    return gpd.GeoDataFrame(gdf_clear, geometry=geo_col, crs=4326)


def fix_self_intersection(geo: (Polygon, MultiPolygon)) -> MultiPolygon:
  """
  修复自相交面数据，通过调用make_valid方法实现，
  将multipolygon拆分成多个polygon处理，
  对于一个自相交的polygon会保留最大的面
  洞结构也能保留下来
  """
  res = MultiPolygon()
  if not geo.is_valid:
    geoms = get_geoms(geo)
    for g in geoms:
      valid = get_geoms(make_valid(g))
      for v in valid:
        if isinstance(v, Polygon):
          res = res.union(v)
    return res
  return res.union(geo)


def topology_check(gdf: gpd.GeoDataFrame,
                   is_fix=True,
                   geo_col='geometry',
                   fill_intersects=False,
                   keep_contains=False) -> gpd.GeoDataFrame:
  """
  完成全部拓扑检查步骤:
  1. 检查geometry是否有空值并报错
  2. 检查面数据是否自相交并修复
  3. 检查面数据是否重叠并修复
  Args:
    fill_intersects: 是否填满相交的区域，如果为True的话，两个面相交的区域会随机分配到其中一个面内
    keep_contains: fill_intersects为True时才有效，即是否保留被包含的面
  """
  gdf = gdf.copy()
  assert_not_null(gdf, geo_col, skip_if_not_exists=False)
  if gdf.duplicated(geo_col).any():
    raise AssertionError('存在重复的geometry')
  if not all(gdf.is_valid):
    logging.warning('面数据:存在自相交板块,%s', gdf[~gdf.is_valid])
    if is_fix:
      gdf[geo_col] = gdf[geo_col].apply(fix_self_intersection)

  err = overlap_check(gdf)
  if not err.empty:
    logging.warning('面数据:存在重叠板块, %s', err)
    if is_fix:
      gdf = fix_overlap(gdf,
                        geo_col,
                        fill_intersects=fill_intersects,
                        keep_contains=keep_contains)
  return gdf



import os

import pandas as pd
from .decorator import print_doc
from .dt import DT


class Docx:
  """Docx"""

  def set_default_style(self):
    from docx import Document
    from docx.oxml.ns import qn
    from docx.shared import Pt
    from docx.shared import RGBColor

    self.doc = Document()
    # 设置默认字体
    self.doc.styles['Normal'].font.name = u'微软雅黑'
    self.doc.styles['Normal']._element.rPr.rFonts.set(
        qn('w:eastAsia'), u'微软雅黑')
    self.doc.styles['Normal'].font.size = Pt(10.5)
    self.doc.styles['Normal'].font.color.rgb = RGBColor(0, 0, 0)
    self.create_time = DT().today

  def add_table_from_df(self, df: pd.DataFrame, newline=False):
    """将dataframe作为表格写入"""
    t = self.doc.add_table(df.shape[0] + 1, df.shape[1], style='Table Grid')
    for j in range(df.shape[-1]):
      t.cell(0, j).text = df.columns[j]
    for i in range(df.shape[0]):
      for j in range(df.shape[-1]):
        t.cell(i + 1, j).text = str(df.values[i, j])
    if newline:
      self.add_normal_p('')

  def add_title(self, text: str, n: int):
    """添加标题"""
    self.doc.add_heading(text, level=n)

  def add_title_0(self, text: str):
    """添加大标题"""
    self.add_title(text, 0)

  def add_title_1(self, text: str):
    """添加一级标题"""
    self.add_title(text, 1)

  def add_title_2(self, text: str):
    """添加二级标题"""
    self.add_title(text, 2)

  def add_title_3(self, text: str):
    """添加三级标题"""
    self.add_title(text, 3)

  def add_title_4(self, text: str):
    """添加四级标题"""
    self.add_title(text, 4)

  def add_normal_p(self, text: str):
    '''添加普通段落'''
    self.doc.add_paragraph(text)

  def add_paragraph_color(self, text: str, rgb: (list, tuple)):
    from docx.shared import RGBColor

    p = self.doc.add_paragraph('')
    p.add_run(text).font.color.rgb = RGBColor(*rgb)

  def add_paragraph_red(self, text: str):
    from docx.shared import RGBColor

    p = self.doc.add_paragraph('')
    p.add_run(text).font.color.rgb = RGBColor(250, 0, 0)

  def add_bullet_list(self, text: str):
    """添加无序列表"""
    self.doc.add_paragraph(text, style='List Bullet')

  def add_order_list(self, text: str):
    """添加有序列表"""
    self.doc.add_paragraph(text, style='List Number')

  def add_intense_quote(self, text: str):
    """添加高亮段落"""
    self.doc.add_paragraph(text, style='Intense Quote')

  @print_doc
  def save(self, file_path: str):
    """保存文件至word文档"""
    self.doc.save(file_path)
    print('文件保存至：', os.path.abspath(file_path))

  def add_hist_from_data(self, data: list, newline=False):
    from docx.shared import Inches
    from matplotlib import pyplot as plt
    plt.figure(figsize=(12, 4))
    plt.style.use('seaborn')
    plt.hist(data)
    plt.savefig('image.png')
    self.doc.add_picture('image.png', width=Inches(6))
    if os.path.exists('image.png'):
      os.remove('image.png')
    if newline:
      self.add_normal_p('')



import csv
import os
import warnings

import geopandas as gpd
import pandas as pd

from ..util.os import ext


def max_grid():
  """防止单个单元格文件过大而报错"""
  import sys
  maxint = sys.maxsize
  decrement = True
  while decrement:
    decrement = False
    try:
      csv.field_size_limit(maxint)
    except OverflowError:
      maxint = int(maxint / 10)
      decrement = True


def rdxls(
    filename,
    sheet_name=0,
    sheet_contains: str = None,
    errors='raise',
    dtype=None,
) -> pd.DataFrame:
  """
  读取excel文件

  :param filename: 文件名
  :param sheet_name: sheet表的名称
  :param sheet_contains: sheet表包含的字符串
  :param errors: 当没有对应sheet时，raise: 抛出错误, coerce: 返回空的dataframe
  :return:
  """
  if sheet_name == 0:
    if sheet_contains is not None:
      df = pd.read_excel(filename, sheet_name=None, dtype=dtype)
      sheet_list = [i for i in df.keys() if sheet_contains in i]
      if len(sheet_list) != 0:
        sheet_name = sheet_list[0]
        if len(sheet_list) == 1:
          print(f"sheet:  <'{sheet_name}'>")
        elif len(sheet_list) >= 2:
          warnings.warn(
              f"包含'{sheet_contains}'的sheet有{sheet_list}，所读取的sheet为:{sheet_name}")
        return df[sheet_name]
      else:
        if errors == 'coerce':
          warnings.warn(f'没有包含{sheet_contains}的sheet，请检查')
          return pd.DataFrame()
        elif errors == 'raise':
          raise ValueError(f'没有包含{sheet_contains}的sheet，请检查')
        else:
          raise KeyError("参数'error'错误, 可选参数为coerce和raise")
  else:
    print(f"sheet:  <'{sheet_name}'>")
  return pd.read_excel(filename, sheet_name=sheet_name, dtype=dtype)


def rdf(
    file_path: str,
    *,
    sheet_name=0,
    sheet_contains: str = None,
    encoding='utf-8-sig',
    info=False,
    dtype=None,
) -> pd.DataFrame:
  """
  常用文件读取函数，支持.csv/.xlsx/.shp/.parquet/.pickle/.feather'
  """
  max_grid()
  ex = ext(file_path)
  if ex == '.csv':
    try:
      df = pd.read_csv(file_path,
                       engine='python',
                       encoding=encoding,
                       dtype=dtype)
    except UnicodeDecodeError:
      df = pd.read_csv(file_path, engine='python', dtype=dtype)
  elif ex in ('.parquet', '.pickle', '.feather'):
    _t = ex.strip('.')
    df = getattr(pd, f'read_{_t}')(
        file_path
    )
  elif ex in ('.xls', '.xlsx'):
    df = rdxls(file_path,
               sheet_name=sheet_name,
               sheet_contains=sheet_contains,
               dtype=dtype)
  elif ex == '.shp':
    try:
      df = gpd.GeoDataFrame.from_file(file_path)
    except UnicodeEncodeError:
      df = gpd.GeoDataFrame.from_file(file_path, encoding='GBK')
  elif ex in ('.kml', '.ovkml'):
    df = read_kml(file_path)
  else:
    raise Exception('未知文件格式')
  if info:
    print(f'shape: {df.shape}')
    print(f'columns: {df.columns}')
  return df


def read_line_json(filename, encoding='utf-8') -> pd.DataFrame:
  """
  逐行读取json格式的文件

  :param filename: 文件名
  :param encoding: 文件编码，默认为utf-8
  :return:
  """
  import json
  records = []
  with open(filename, 'r', encoding=encoding) as file:
    for line in file:
      row = json.loads(line)
      records.append(row)
  return pd.DataFrame(records)


def bigdata2df(filename: str,
               chunksize: int = 10000,
               code: str = 'utf-8') -> pd.DataFrame:
  reader = pd.read_table(filename,
                         encoding=code,
                         sep=',',
                         skip_blank_lines=True,
                         iterator=True)
  loop = True
  chunks = []
  while loop:
    try:
      chunk = reader.get_chunk(chunksize)
      chunk.dropna(axis=0, inplace=True)
      chunks.append(chunk)
    except StopIteration:
      loop = False
      print('Iteration is stopped.')
  df = pd.concat(chunks, ignore_index=True, axis=1)
  return df


def read_parquet_by_dir(dir_path):
  df = pd.DataFrame()
  for filename in os.listdir(dir_path):
    if ext(filename) == '.parquet':
      df = pd.concat([df, rdf(os.path.join(dir_path, filename))])
  return df


def kml_df_create_level(gdf_dict) -> dict:
  """
  读取 dict {level: gpd.GeoDataFrame} 字典, 并根据[name]列包含关系提取
  分辨文件夹层级
  Args:
    gdf_dict: {dict} {level: gpd.GeoDataFrame} key为文件夹名称, value为对应数据
  Returns: dict: 文件夹层级字典
      example:
        {'板块名称': 0,
        'T顶豪': 1,
        'TOP1城区顶级': 2,
        'O远郊': 1,
        'O1远郊品质': 2,
        'O2远郊安居': 2}
  """
  level_dict = {}
  for i in gdf_dict.keys():
    level_dict[i] = 0
  for k in range(1, len(gdf_dict.keys())):
    for j in list(gdf_dict.keys())[k:]:
      sample = list(gdf_dict.keys())[k - 1]
      if set(gdf_dict[j]['name']) <= set(gdf_dict[sample]['name']):
        level_dict[j] = level_dict[sample] + 1
  return level_dict


def get_kml_df_with_level(gdf_dict) -> gpd.GeoDataFrame:
  """
  从中gdf_dict读取数据以及文件夹名称,并构造文件夹层级,按照指定层级合并所
  有数据,并打上层级标签
  Args:
    gdf_dict{dict}:{level: gpd.GeoDataFrame} key为文件夹名称, value为对应数据
  """
  level_dict = kml_df_create_level(gdf_dict)
  level_dict_new = {}
  for k, v in level_dict.items():
    columns_name = f'level_{v}'
    if columns_name not in level_dict_new.keys():
      level_dict_new[columns_name] = []
    gdf_dict[k][columns_name] = k
    level_dict_new[columns_name].append(gdf_dict[k])
  data_all = gpd.GeoDataFrame(columns=['name', 'geometry'])
  for v in level_dict_new.values():
    level_df = pd.concat(v)
    data_all = data_all.merge(level_df, on=['name', 'geometry'], how='outer')
  data_all['name'] = data_all['name'].replace('-', None)
  return data_all


def read_kml(file_path, keep_level=True) -> gpd.GeoDataFrame:
  """
  读取kml类文件,将其转换成DataFrame, 根据separate_folders选择是否拆分,
  输出仅保留[name, geometry]字段,暂时只支持polygon和multipolygon数据
  Args:
    file_path{str}: kml or ovkml文件路径
    keep_level(bool): 是否保留文件夹层级标签
  """
  import kml2geojson as k2g
  features = k2g.convert(file_path, separate_folders=keep_level)
  gdf_dict = {}
  if keep_level:
    for i in range(len(features)):
      gdf = gpd.GeoDataFrame.from_features(features[i]['features'])
      level = features[i]['name']
      if 'name' not in gdf.columns:
        gdf['name'] = '-'
      gdf = gdf[['name', 'geometry']]
      gdf_dict[level] = gdf
    data_all = get_kml_df_with_level(gdf_dict)
  else:
    data_all = gpd.GeoDataFrame.from_features(features[0]['features'])
    data_all = data_all[['name', 'geometry']]
  return data_all



import warnings

from ..geometry.df import lnglat2shapely as geom_lnglat2shapely
from ..geometry.df import lnglat2wkb as geom_lnglat2wkb
from ..geometry.df import lnglat2wkt as geom_lnglat2wkt
from ..geometry.df import mark_tags_v2
from ..geometry.df import nearest_neighbor
from ..geometry.df import projection
from ..geometry.df import buffer
from ..geometry.df import spatial_agg
from ..geometry.df import auto2shapely as ensure_gdf
from ..geometry.df import shapely2lnglat as geom_shapely2lnglat
from ..geometry.df import shapely2wkb as geom_shapely2wkb
from ..geometry.df import shapely2wkt as geom_shapely2wkt
from ..geometry.df import shapely2x as geom_shapely_to
from ..geometry.df import wkb2lnglat as geom_wkb2lnglat
from ..geometry.df import wkb2shapely as geom_wkb2shapely
from ..geometry.df import wkb2wkt as geom_wkb2wkt
from ..geometry.df import wkt2lnglat as geom_wkt2lnglat
from ..geometry.df import wkt2shapely as geom_wkt2shapely
from ..geometry.df import wkt2wkb as geom_wkt2wkb
from ..geometry.util import crs_sh2000
from ..geometry.util import ensure_multi_geom
from ..geometry.util import distance
from ..geometry.util import get_epsg
from ..geometry.util import get_epsg_by_lng
from ..geometry.util import get_inner_point
from ..geometry.util import get_lng_by_city
from ..geometry.util import infer_geom_format
from ..geometry.util import is_geojson
from ..geometry.util import is_shapely
from ..geometry.util import is_wkb
from ..geometry.util import is_wkt
from ..geometry.util import multiline2multipolygon
from ..geometry.util import projection_lnglat
from ..geometry.util import wkb_dumps
from ..geometry.util import wkb_loads
from ..geometry.util import wkt_dumps
from ..geometry.util import wkt_loads

warnings.warn(
    'util.geom模块不再更新，请使用"ricco.geometry"中的相关模块',
    DeprecationWarning
)



import logging
import warnings
from typing import List
from typing import Union

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Point
from shapely.geometry import Polygon
from tqdm import tqdm

from ..util.decorator import geom_progress
from ..util.util import ensure_list
from ..util.util import first_notnull_value
from ..util.util import not_empty
from .util import geojson_dumps
from .util import geojson_loads
from .util import get_epsg
from .util import get_epsg_by_lng
from .util import infer_geom_format
from .util import wkb_dumps
from .util import wkb_loads
from .util import wkt_dumps
from .util import wkt_loads


def projection(
    gdf: gpd.GeoDataFrame,
    epsg: int = None,
    city: str = None,
    geometry='geometry'):
  """投影变换"""
  if not epsg:
    epsg = get_epsg(city) if city else get_epsg_by_lng(
        gdf[geometry].centroid.x.tolist()
    )
  return gdf.to_crs(epsg=epsg)


@geom_progress
def wkb2shapely(df, geometry='geometry',
                epsg_code: int = 4326) -> gpd.GeoDataFrame:
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkb_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@geom_progress
def shapely2wkb(df, geometry='geometry'):
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkb_dumps)
  return df


@geom_progress
def wkt2shapely(df, geometry='geometry',
                epsg_code: int = 4326) -> gpd.GeoDataFrame:
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkt_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@geom_progress
def shapely2wkt(df, geometry='geometry'):
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkt_dumps)
  return df


@geom_progress
def geojson2shapely(df, geometry='geometry',
                    epsg_code: int = 4326) -> gpd.GeoDataFrame:
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(geojson_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@geom_progress
def shapely2geojson(df, geometry='geometry'):
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(geojson_dumps)
  return df


@geom_progress
def lnglat2shapely(df,
                   lng='lng',
                   lat='lat',
                   geometry='geometry',
                   delete=True,
                   epsg_code: int = 4326) -> gpd.GeoDataFrame:
  df = df.copy()
  df[geometry] = df.progress_apply(
      lambda d: Point(d[lng], d[lat])
      if not_empty(d[lng]) and not_empty(d[lat])
      else None,
      axis=1
  )
  df = gpd.GeoDataFrame(df, crs=epsg_code, geometry=geometry)
  if delete:
    del df[lng], df[lat]
  return df


@geom_progress
def shapely2lnglat(df, geometry='geometry',
                   lng='lng', lat='lat',
                   within=False, delete=False):
  """
  shapely格式提取面内点转为经纬度。
  Args:
    within: 范围的点是否再面内
      - False(default): 直接返回中心点；
      - True: 返回面内的一个点
  """

  df = df.copy()
  if within:
    df[lng] = df.representative_point().x
    df[lat] = df.representative_point().y
  else:
    df[lng] = df.centroid.x
    df[lat] = df.centroid.y
  if delete:
    del df[geometry]
  return df


def wkb2lnglat(df, geometry='geometry', delete=False, within=False):
  """geometry转经纬度，求中心点经纬度"""
  df = wkb2shapely(df, geometry=geometry)
  df = shapely2lnglat(df, geometry=geometry, within=within, delete=delete)
  if not delete:
    df = shapely2wkb(df, geometry=geometry)
  return df


def lnglat2wkb(df,
               lng='lng',
               lat='lat',
               geometry='geometry',
               delete=False, code=4326):
  """经纬度转wkb格式的geometry"""
  df = lnglat2shapely(
      df, lng, lat, geometry=geometry, delete=delete, epsg_code=code
  )
  df = shapely2wkb(df, geometry=geometry)
  return df


def wkt2lnglat(df, geometry='geometry', delete=False, within=False):
  """geometry转经纬度，求中心点经纬度"""
  df = wkt2shapely(df, geometry=geometry)
  df = shapely2lnglat(df, geometry=geometry, within=within, delete=delete)
  if not delete:
    df = shapely2wkt(df, geometry=geometry)
  return df


def lnglat2wkt(df,
               lng='lng',
               lat='lat',
               geometry='geometry',
               delete=False, code=4326):
  """经纬度转wkb格式的geometry"""
  df = lnglat2shapely(
      df, lng, lat, geometry=geometry, delete=delete, epsg_code=code
  )
  df = shapely2wkt(df, geometry=geometry)
  return df


def wkb2wkt(df, geometry='geometry', epsg_code: int = 4326):
  """wkb转wkt"""
  df = wkb2shapely(df, geometry=geometry, epsg_code=epsg_code)
  return shapely2wkt(df, geometry=geometry)


def wkt2wkb(df, geometry='geometry', epsg_code: int = 4326):
  """wkb转wkt"""
  df = wkt2shapely(df, geometry=geometry, epsg_code=epsg_code)
  return shapely2wkb(df, geometry=geometry)


def shapely2central_shapely(df, geometry='geometry', within=False):
  df = shapely2lnglat(df, geometry=geometry, within=within)
  return lnglat2shapely(df, geometry=geometry, delete=True)


def auto2shapely(df, geometry='geometry'):
  _geom = first_notnull_value(df[geometry])
  geom_format = infer_geom_format(_geom)
  if geom_format == 'geojson':
    return geojson2shapely(df, geometry=geometry)
  if geom_format == 'wkb':
    return wkb2shapely(df, geometry=geometry)
  if geom_format == 'wkt':
    return wkt2shapely(df, geometry=geometry)
  if geom_format == 'shapely':
    return gpd.GeoDataFrame(df, geometry=geometry)
  raise TypeError('未知的地理格式，支持wkb,wkt,shapely三种格式')


def shapely2x(df, geometry_format: str, geometry='geometry'):
  """
  将shapely转为指定的格式
  Args:
    df: 要转换的Dataframe
    geometry_format: 支持wkb,wkt,shapely,geojson
    geometry: geometry列的列名，默认“geometry”
  """
  if geometry_format == 'geojson':
    return shapely2geojson(df, geometry=geometry)
  if geometry_format == 'wkb':
    return shapely2wkb(df, geometry=geometry)
  if geometry_format == 'wkt':
    return shapely2wkt(df, geometry=geometry)
  if geometry_format == 'shapely':
    return df
  raise ValueError('未知的地理格式，支持wkb,wkt,shapely三种格式')


def auto2x(df, geometry_format: str, geometry='geometry'):
  """
  将geometry转为指定格式
  Args:
    df: 要转换的Dataframe
    geometry_format: 要转换为的geometry类型，支持shapely,wkb,wkt,geojson
    geometry: geometry列的列名，默认为“geometry”
  """
  geom = first_notnull_value(df[geometry])
  __geom_format = infer_geom_format(geom)
  if __geom_format == geometry_format:
    return df
  df = auto2shapely(df, geometry=geometry)
  return shapely2x(df, geometry_format=geometry_format, geometry=geometry)


def distance_min(geometry, gdf: gpd.GeoDataFrame) -> float:
  """
  计算单个geometry到数据集gdf中元素的最短距离
  Args:
    geometry: 单个geometry，shapely格式
    gdf: 数据集，GeoDataFrame格式
  """
  return gdf.distance(geometry).min()


@geom_progress
def distance_gdf(df: gpd.GeoDataFrame, df_target: gpd.GeoDataFrame, c_dst: str,
                 left_geometry='geometry'):
  """计算一个数据集中的每个元素到另一个数据集之间的最短距离"""
  df = df.copy()
  df[c_dst] = df[left_geometry].progress_apply(
      lambda p: distance_min(p, df_target) if not_empty(p) else np.nan
  )
  return df


def split_grids(df: gpd.GeoDataFrame, step: int, city: str = None):
  """
  根据所给边界划分固定边长的栅格

  Args:
      df: 边界文件，GeoDataFrame格式
      step: 栅格边长，单位：米
      city: 所属城市，用于投影
  """

  def get_xxyy(df):
    bounds_dict = df.bounds.T.to_dict()[0]
    minx = bounds_dict['minx']
    miny = bounds_dict['miny']
    maxx = bounds_dict['maxx']
    maxy = bounds_dict['maxy']
    xx = maxx - minx
    yy = maxy - miny
    return [xx, yy, minx, miny, maxx, maxy]

  def get_lnglat(df, i, j):
    _lng = df.loc[(df['i'] == i) &
                  (df['j'] == j), 'lng'].reset_index(drop=True)[0]
    _lat = df.loc[(df['i'] == i) &
                  (df['j'] == j), 'lat'].reset_index(drop=True)[0]
    return (_lng, _lat)

  def get_lnglat_sets(df, i, j):
    A = get_lnglat(df, i, j)
    B = get_lnglat(df, i, j + 1)
    C = get_lnglat(df, i + 1, j + 1)
    D = get_lnglat(df, i + 1, j)
    return Polygon((A, B, C, D, A))

  # 融合
  if not isinstance(df, gpd.GeoDataFrame):
    df = wkb2shapely(df)
  df = df[['geometry']].dissolve()

  # 投影，获取实际的距离
  df_p = projection(df, city=city)

  # 分别获取投影前和投影后的点位及边长信息
  xy_o, xy_p = get_xxyy(df), get_xxyy(df_p)

  # 获取栅格的起止经纬度，步长等信息
  xr, yr = xy_o[0] / xy_p[0], xy_o[1] / xy_p[1]
  x_len, y_len = xr * step, yr * step
  x_num, y_num = int(xy_o[0] // x_len + 1), int(xy_o[1] // y_len + 1)
  logging.info(f'lng分段数：{x_num}，lat分段数：{y_num}')
  x_start, y_start = xy_o[2], xy_o[3]
  x_end, y_end = x_start + x_num * x_len, y_start + y_num * y_len

  # 构建经纬度列表
  lng_array, lat_array = np.linspace(x_start, x_end, num=x_num), np.linspace(
      y_start, y_end, num=y_num)
  lng_list, lat_list = list(lng_array), list(lat_array)

  # 构建栅格索引
  i_l, j_l, lng_l, lat_l = [], [], [], []
  for i, lng in enumerate(lng_list):
    for j, lat in enumerate(lat_list):
      i_l.append(i)
      j_l.append(j)
      lng_l.append(lng)
      lat_l.append(lat)
  df_temp = pd.DataFrame({'i': i_l, 'j': j_l, 'lng': lng_l, 'lat': lat_l})
  # 删除末尾的格子，避免后面出现out of range
  df_res = df_temp.loc[
    (df_temp['i'] != x_num - 1) & (df_temp['j'] != y_num - 1),
    ['i', 'j']].reset_index(drop=True)

  # 组合坐标集并转为Polygon
  tqdm.pandas(desc='lnglats2shapelyPolygon')
  df_res['geometry'] = df_res.progress_apply(
      lambda df: get_lnglat_sets(df_temp, df['i'], df['j']), axis=1)

  # 生成grid_id列
  df_res['grid_id'] = df_res['i'].astype(str) + '-' + df_res['j'].astype(str)
  del df_res['i'], df_res['j']

  # 通过边界裁剪栅格
  df_res = gpd.GeoDataFrame(df_res, crs="epsg:4326")
  df.crs = 'epsg:4326'
  df_res = gpd.sjoin(df_res,
                     df,
                     how='inner',
                     predicate='intersects').drop('index_right', axis=1)

  # 将geometry转为wkb格式
  df_res = shapely2wkb(df_res)
  return df_res


def _ensure_geometry(_df,
                     ensure_point=False,
                     warning_message=False,
                     geometry='geometry',
                     lng='lng', lat='lat'):
  """转换并仅保留geometry列"""
  if geometry in _df:
    _df = auto2shapely(_df[[geometry]], geometry=geometry)[[geometry]]
    if ensure_point:
      geom = first_notnull_value(_df[geometry])
      if geom.geom_type not in ['point', 'Point']:
        if warning_message:
          warnings.warn('非点数据，将自动提取中心点')
        _df = shapely2central_shapely(_df, within=True, geometry=geometry)
    return _df
  elif lng in _df and lat in _df:
    return lnglat2shapely(
        _df[[lng, lat]],
        lng=lng, lat=lat,
        geometry=geometry,
        delete=False,
    )[[geometry]]
  else:
    raise KeyError(f'文件中必须有"lng"+"lat"列或"{geometry}"列')


def mark_tags_v2(
    point_df: pd.DataFrame,
    polygon_df: pd.DataFrame,
    col_list: list = None,
    predicate='intersects',
    drop_geometry=False,
    geometry_format='wkb',
    warning_message=True,
    point_lng='lng',
    point_lat='lat',
    point_geometry='geometry',
    polygon_geometry='geometry',
):
  """
  使用面数据通过空间关联（sjoin）给数据打标签

  Args:
      point_df: 点数据
      polygon_df: 面数据
      col_list: 面数据中要关联到结果中的列，若为空则全部关联
      predicate: 关联方法，默认'intersects'
      drop_geometry: 结果是否删除geometry，默认删除
      geometry_format: 输出的geometry格式，支持wkb,wkt,shapely,geojson，默认wkb
      warning_message: 是否输出警告信息
  """
  point_df = point_df.copy()
  if not col_list:
    col_list = polygon_df.columns.to_list()
  else:
    col_list = ensure_list(col_list)
    polygon_df = polygon_df[[*col_list, polygon_geometry]]

  for c in col_list:
    if c in point_df and c not in [point_lng, point_lat, point_geometry]:
      c_n = f'{c}_origin'
      if warning_message:
        warnings.warn(f'点数据中存在面文件中待关联的列，已重命名：{c} --> {c_n}')
      point_df.rename(columns={c: c_n}, inplace=True)
  # 转换为shapely格式
  df = _ensure_geometry(point_df, True, warning_message,
                        lng=point_lng, lat=point_lat, geometry=point_geometry)
  polygon_df = auto2shapely(polygon_df, geometry=polygon_geometry)
  # 空间关联
  df = df.sjoin(
      polygon_df, how='left', predicate=predicate,
  ).drop(['index_right'], axis=1)
  # 统一geometry输出格式、删除geometry、避免多次转换
  if point_geometry in point_df:
    if geometry_format != 'shapely':
      del df[point_geometry]
    if geometry_format == 'shapely' or drop_geometry:
      del point_df[point_geometry]
    else:
      point_df = auto2x(point_df, geometry_format, geometry=point_geometry)
  elif drop_geometry:
    del df[point_geometry]
  else:
    df = shapely2x(df, geometry_format, geometry=point_geometry)
  # 将空间关联后的数据关联到原来的Dataframe上
  return point_df.join(df, how='left')


def nearest_neighbor(
    df: pd.DataFrame,
    df_target: pd.DataFrame,
    c_dst='min_distance',
    epsg: int = None) -> pd.DataFrame:
  """
  近邻分析，计算一个数据集中的元素到另一个数据集中全部元素的最短距离（单位：米）

  Args:
    df:
    df_target:
    c_dst: 输出最短距离的列名，默认为“min_distance”
    epsg: 对于跨时区或不在同一个城市的可以指定epsg code，默认会根据经度中位数获取
  """
  # 将两个数据集都转为shapely格式
  df_left = _ensure_geometry(df)
  df_target = _ensure_geometry(df_target)
  # 投影
  df_left = projection(df_left, epsg=epsg)
  df_target = projection(df_target, epsg=epsg)
  # 计算最短距离
  df_left = distance_gdf(df_left, df_target, c_dst)
  # 将距离合并到原来的数据集上
  return df.join(df_left[[c_dst]], how='left')


def get_area(
    df: pd.DataFrame,
    c_dst='area',
    epsg: int = None) -> pd.DataFrame:
  """
  计算面积（单位：平方米）
  Args:
    df: 要计算的面数据
    c_dst: 输出面积的列名，默认为“area”
    epsg: 对于跨时区或不在同一个城市的可以指定epsg code，默认会根据经度中位数获取
  """
  # 将数据集转为shapely格式
  df_left = _ensure_geometry(df)
  # 投影
  df_left = projection(df_left, epsg=epsg)
  # 计算面积
  df_left[c_dst] = df_left.area
  # 将面积合并到原来的数据集上
  return df.join(df_left[[c_dst]], how='left')


def buffer(df: pd.DataFrame, radius: Union[int, float],
           city: str = None,
           geo_type: str = 'point',
           geometry: str = 'geometry',
           buffer_geometry: str = 'buffer_geometry',
           geo_format='wkb') -> pd.DataFrame:
  """
  获得一定半径的缓冲区

  Args:
    df: pd.DataFrame, 包含地理信息的DataFrame
    radius: numeric, 缓冲区半径（单位米）
    city: str, 可选, 投影城市，可提高数据精度
    geo_type: str, 地理数据类型，可选point, line或polygon(包括multipolygon)，默认point
    geometry: str, geometry字段名，默认"geometry"
    buffer_geometry: 输出的缓冲区geometry字段名，默认"buffer_geometry"
    geo_format: str, 输出的缓冲区geometry格式，支持wkb,wkt,shapely,geojson，默认wkb

  Returns: 包含缓冲区geometry的DataFrame

  Examples:
    >>> df = pd.DataFrame({'id': [1, 2, 3],
    >>>                    'lng': [116.18601, 116.18366, 116.18529],
    >>>                    'lat': [40.02894, 40.03550, 40.03565]})

    >>> df
        id        lng       lat
    0   1  116.18601  40.02894
    1   2  116.18366  40.03550
    2   3  116.18529  40.03565

    >>> buffer(df=df, radius=1500, city='北京', geo_type='point')
        id        lng       lat                                    buffer_geometry
    0   1  116.18601  40.02894  0103000020E6100000010000004100000092A0207A070D...
    1   2  116.18366  40.03550  0103000020E6100000010000004100000071178800E10C...
    2   3  116.18529  40.03565  0103000020E610000001000000410000008A2570B5FB0C...

  """
  df = df.copy()
  if geo_type == 'point':
    df_buffer = _ensure_geometry(df, True)
  elif geo_type in ['line', 'polygon']:
    df_buffer = auto2shapely(df, geometry=geometry)[[geometry]]
  else:
    raise ValueError('geo_type必须为point，line或polygon')

  df_buffer = projection(df_buffer, city=city)
  df_buffer[buffer_geometry] = df_buffer.buffer(radius)
  df_buffer = gpd.GeoDataFrame(
      df_buffer[[buffer_geometry]], geometry=buffer_geometry
  )
  df_buffer = projection(df_buffer, epsg=4326, geometry=buffer_geometry)
  df_buffer = shapely2x(df_buffer, geo_format)
  return df.join(df_buffer, how='left')


def spatial_agg(point_df: pd.DataFrame, polygon_df: pd.DataFrame,
                by: Union[str, List[str]],
                agg: dict,
                polygon_geometry: str = 'geometry') -> pd.DataFrame:
  """
  对面数据覆盖范围内的点数据进行空间统计
  Args:
    point_df: pd.DataFrame, 点数据dataframe;
    polygon_df: pd.DataFrame, 面数据dataframe;
    by: Union[str, List[str]], 空间统计单位字段；
    agg: dict, 空间统计操作。格式为{'被统计字段名': '操作名', ...}的字典。如{'poi':'sum'};
    polygon_geometry: str, 面数据geometry字段名，默认"geometry";

  Returns: pd.DataFrame, 包含空间统计单位字段和被统计字段和面数据geometry的DataFrame
  """

  polygon_df = polygon_df[[by, polygon_geometry]]
  point_df = mark_tags_v2(polygon_df=polygon_df, point_df=point_df,
                          drop_geometry=True, polygon_geometry=polygon_geometry)
  df_grouped = point_df.groupby(by=by, as_index=False).agg(agg)
  return df_grouped



import math
from enum import Enum
from functools import partial

import geojson.utils
import pandas as pd
import pyproj
from shapely import geos
from shapely import ops
from shapely import wkb
from shapely import wkt
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.geometry.base import BaseGeometry
from shapely.geometry.base import BaseMultipartGeometry
from shapely.ops import transform as sh_transform

"""
Dependencies:

pyproj==2.3.1
pandas==0.24.2
shapely[vectorized]==1.7a1
geojson==2.5.0
"""

__all__ = ['wgs2gcj', 'gcj2wgs', 'gcj2wgs_exact', 'distance', 'gcj2bd',
           'bd2gcj', 'wgs2bd', 'bd2wgs']

earthR = 6378245.0
x_pi = math.pi * 3000.0 / 180.0
LNG_CHINA_RANGE = (72.004, 137.8347)
LAT_CHINA_RANGE = (0.8293, 55.8271)
LNG_RANGE = (-180, 180)
LAT_RANGE = (-90, 90)


class SRS(Enum):
  """空间参考系统(Spatial Reference System)"""
  #: 世界大地测量系统 (World Geodetic System 1984)
  wgs84 = 'wgs84'
  #: 百度坐标系
  bd09 = 'bd09'
  #: 国测局坐标（或火星坐标）
  gcj02 = 'gcj02'


def isnull(x):
  """determine whether input value is null, unlike pandas.isnull, this
  function only handles scalar value.

  null value includes None, pandas NaN/NaT
  """
  # pd.isnull will return array/list/frame for array/list/dataframe input, for
  # all those input, we considered is as not null
  result = pd.isnull(x)
  if not isinstance(result, bool):
    return False
  return result


def outOfChina(lat, lng):
  return not (72.004 <= lng <= 137.8347 and 0.8293 <= lat <= 55.8271)


def transform(x, y):
  xy = x * y
  absX = math.sqrt(abs(x))
  xPi = x * math.pi
  yPi = y * math.pi
  d = 20.0 * math.sin(6.0 * xPi) + 20.0 * math.sin(2.0 * xPi)

  lat = d
  lng = d

  lat += 20.0 * math.sin(yPi) + 40.0 * math.sin(yPi / 3.0)
  lng += 20.0 * math.sin(xPi) + 40.0 * math.sin(xPi / 3.0)

  lat += 160.0 * math.sin(yPi / 12.0) + 320 * math.sin(yPi / 30.0)
  lng += 150.0 * math.sin(xPi / 12.0) + 300.0 * math.sin(xPi / 30.0)

  lat *= 2.0 / 3.0
  lng *= 2.0 / 3.0

  lat += -100.0 + 2.0 * x + 3.0 * y + 0.2 * y * y + 0.1 * xy + 0.2 * absX
  lng += 300.0 + x + 2.0 * y + 0.1 * x * x + 0.1 * xy + 0.1 * absX

  return lat, lng


def delta(lat, lng):
  ee = 0.00669342162296594323
  dLat, dLng = transform(lng - 105.0, lat - 35.0)
  radLat = lat / 180.0 * math.pi
  magic = math.sin(radLat)
  magic = 1 - ee * magic * magic
  sqrtMagic = math.sqrt(magic)
  dLat = (dLat * 180.0) / ((earthR * (1 - ee)) / (magic * sqrtMagic) * math.pi)
  dLng = (dLng * 180.0) / (earthR / sqrtMagic * math.cos(radLat) * math.pi)
  return dLat, dLng


def wgs2gcj(wgsLat, wgsLng):
  if outOfChina(wgsLat, wgsLng):
    return wgsLat, wgsLng
  dlat, dlng = delta(wgsLat, wgsLng)
  return wgsLat + dlat, wgsLng + dlng


def gcj2wgs(gcjLat, gcjLng):
  if outOfChina(gcjLat, gcjLng):
    return gcjLat, gcjLng
  dlat, dlng = delta(gcjLat, gcjLng)
  return gcjLat - dlat, gcjLng - dlng


def gcj2wgs_exact(gcjLat, gcjLng):
  initDelta = 0.01
  threshold = 0.000001
  dLat = dLng = initDelta
  mLat = gcjLat - dLat
  mLng = gcjLng - dLng
  pLat = gcjLat + dLat
  pLng = gcjLng + dLng
  for _ in range(30):
    wgsLat = (mLat + pLat) / 2
    wgsLng = (mLng + pLng) / 2
    tmplat, tmplng = wgs2gcj(wgsLat, wgsLng)
    dLat = tmplat - gcjLat
    dLng = tmplng - gcjLng
    if abs(dLat) < threshold and abs(dLng) < threshold:
      return wgsLat, wgsLng
    if dLat > 0:
      pLat = wgsLat
    else:
      mLat = wgsLat
    if dLng > 0:
      pLng = wgsLng
    else:
      mLng = wgsLng
  return wgsLat, wgsLng


def distance(latA, lngA, latB, lngB):
  pi180 = math.pi / 180
  arcLatA = latA * pi180
  arcLatB = latB * pi180
  x = (math.cos(arcLatA) * math.cos(arcLatB)
       * math.cos((lngA - lngB) * pi180))
  y = math.sin(arcLatA) * math.sin(arcLatB)
  s = x + y
  if s > 1:
    s = 1
  if s < -1:
    s = -1
  alpha = math.acos(s)
  return alpha * earthR


def gcj2bd(gcjLat, gcjLng):
  if outOfChina(gcjLat, gcjLng):
    return gcjLat, gcjLng
  x = gcjLng
  y = gcjLat
  z = math.hypot(x, y) + 0.00002 * math.sin(y * x_pi)
  theta = math.atan2(y, x) + 0.000003 * math.cos(x * x_pi)
  bdLng = z * math.cos(theta) + 0.0065
  bdLat = z * math.sin(theta) + 0.006
  return bdLat, bdLng


def bd2gcj(bdLat, bdLng):
  if outOfChina(bdLat, bdLng):
    return bdLat, bdLng
  x = bdLng - 0.0065
  y = bdLat - 0.006
  z = math.hypot(x, y) - 0.00002 * math.sin(y * x_pi)
  theta = math.atan2(y, x) - 0.000003 * math.cos(x * x_pi)
  gcjLng = z * math.cos(theta)
  gcjLat = z * math.sin(theta)
  return gcjLat, gcjLng


def wgs2bd(wgsLat, wgsLng):
  return gcj2bd(*wgs2gcj(wgsLat, wgsLng))


def bd2wgs(bdLat, bdLng):
  return gcj2wgs(*bd2gcj(bdLat, bdLng))


_project_to_meter = partial(
    pyproj.transform,
    pyproj.Proj('epsg:4326'),
    pyproj.Proj('epsg:3857'))


def project_to_meter(geometry):
  """transform epsg:4326 to epsg:3857"""
  return ops.transform(_project_to_meter, geometry)


_project_to_lnglat = partial(
    pyproj.transform,
    pyproj.Proj('epsg:3857'),
    pyproj.Proj('epsg:4326'))


def project_to_lnglat(geometry):
  """transform epsg:3857 to epsg:4326"""
  return ops.transform(_project_to_lnglat, geometry)


def cut(line, dist):
  # Cuts a line in two at a distance from its starting point
  if dist <= 0.0 or dist >= line.length:
    return [LineString(line)]
  coords = list(line.coords)
  for i, p in enumerate(coords):
    pdist = line.project(Point(p))
    if pdist == dist:
      return [
        LineString(coords[:i + 1]),
        LineString(coords[i:])]
    if pdist > dist:
      cp = line.interpolate(dist)
      return [
        LineString(coords[:i] + [(cp.x, cp.y)]),
        LineString([(cp.x, cp.y)] + coords[i:])]


geos.WKBWriter.defaults['include_srid'] = True


def wkt_2_ewkb(geometry, srid=4326):
  geometry = wkt.loads(geometry)
  geos.lgeos.GEOSSetSRID(geometry._geom, srid)
  return geometry.wkb_hex


def wkb_2_ewkb(geometry, srid=4326):
  geometry = wkb.loads(geometry, hex=True)
  geos.lgeos.GEOSSetSRID(geometry._geom, srid)
  return geometry.wkb_hex


def geometry_2_ewkb(geometry, srid=4326):
  geos.lgeos.GEOSSetSRID(geometry._geom, srid)
  return geometry.wkb_hex


_fn_mapping = {
  (SRS.bd09, SRS.wgs84): bd2wgs,
  (SRS.gcj02, SRS.wgs84): gcj2wgs,
  (SRS.wgs84, SRS.bd09): wgs2bd,
  (SRS.gcj02, SRS.bd09): gcj2bd,
  (SRS.wgs84, SRS.gcj02): wgs2gcj,
  (SRS.bd09, SRS.gcj02): bd2gcj,
}


def coord_transform(lng: float, lat: float, from_srs: (SRS, str),
                    to_srs: (SRS, str)):
  """坐标系转换

  :param lng: 输入的经度
  :param lat: 输入的纬度
  :param from_srs: 输入坐标的格式
  :param to_srs: 输出坐标的格式
  """
  if from_srs == to_srs:
    return lng, lat

  if isnull(lng) or isnull(lat):
    return None, None

  key = (from_srs, to_srs)
  if key not in _fn_mapping:
    raise NotImplementedError(
        'not support transformation from %s to %s' % (from_srs, to_srs))
  lat, lng = _fn_mapping[key](lat, lng)
  return lng, lat


def coord_transform_geojson(obj: dict, from_srs: SRS, to_srs: SRS):
  """对GeoJSON格式内的所有坐标点执行坐标系转换

  Usage:
    from shapely.geometry import mapping
    from shapely.geometry import Point
    from shapely.geometry import shape

    point_bd09 = Point([1, 2])
    geojson = mapping(point_bd09)
    transformed = coord_transform(geojson, SRS.bd09, SRS.wgs84)
    point_wgs84 = shape(transformed)

  """
  return geojson.utils.map_tuples(
      lambda c: coord_transform(c[0], c[1], from_srs, to_srs), obj)


def coord_transform_geometry(geo: (BaseGeometry, BaseMultipartGeometry),
                             from_srs: SRS,
                             to_srs: SRS):
  """对Geomery内的所有点进行坐标转换，返回转换后的Geometry

  该方法可以支持所有的Shapely Geometry形状，包括Point, Line, Polygon,
  MultiPloygon等，返回的Geometry和输入的形状保持一致
  Args:
    geo: 输入的shapely Geometry
    from_srs: 输入的坐标格式
    to_srs: 输出的坐标格式
  Returns:
    转换后的shapely Geometry
  """
  return sh_transform(
      lambda x, y, z=None: coord_transform(x, y, from_srs, to_srs), geo)


def lnglat_check(df):
  """检查列名"""
  if ('lng' not in df.columns) | ('lng' not in df.columns):
    raise KeyError('经纬度列名必须为lng和lat')


def coord_trans_x2y(df_org, srs_from: (SRS, str), srs_to: (SRS, str)):
  """
  坐标批量转换工具

  :param df_org: 输入的dataframe，经纬度为lng和lat
  :param srs_from: 当前坐标系，可选'wgs84', 'bd09', 'gcj02'
  :param srs_to: 要转的坐标系，可选'wgs84', 'bd09', 'gcj02'
  :return:
  """

  def fn(row):
    return coord_transform(row['lng'], row['lat'], srs_from, srs_to)

  lnglat_check(df_org)
  df_org['lng'] = df_org['lng'].astype(float)
  df_org['lat'] = df_org['lat'].astype(float)
  df_org[['lng', 'lat']] = df_org.apply(fn, axis=1, result_type='expand')
  print('坐标转换完成：%s → %s' % (srs_from, srs_to))
  return df_org


def BD2WGS(df_org):
  """百度转WGS84"""
  df_org = coord_trans_x2y(df_org, SRS.bd09, SRS.wgs84)
  return df_org


def GD2WGS(df_org):
  """高德转WGS84"""
  df_org = coord_trans_x2y(df_org, SRS.gcj02, SRS.wgs84)
  return df_org


def coord_trans_geom(df_org, srs_from: (SRS, str), srs_to: (SRS, str)):
  """
  坐标批量转换工具geometry

  :param df_org: 输入的dataframe，必须要有geometry列
  :param srs_from: 当前坐标系，可选'wgs84', 'bd09', 'gcj02'
  :param srs_to: 要转的坐标系，可选'wgs84', 'bd09', 'gcj02'
  :return:
  """
  from ..geometry.util import wkb_dumps
  from ..geometry.util import wkb_loads
  if 'geometry' not in df_org.columns:
    raise KeyError('必须有geometry列')
  df_org['geometry'] = df_org['geometry'].apply(
      lambda x: wkb_dumps(
          coord_transform_geometry(wkb_loads(x, hex=True),
                                   srs_from,
                                   srs_to),
          hex=True,
          srid=4326))
  print(f'坐标转换：{srs_from} --> {srs_to}')
  return df_org



import os

from ..util.os import fn
from .extract import rdf
from .transformer import df_iter


def split2csvs(
    filename: str,
    *,
    chunksize: int = None,
    parts: int = None,
):
  """将文件拆分为多个文件，放置在与文件同名文件夹下"""
  if not any([chunksize, parts]):
    raise ValueError(f'chunksize 和 parts必须指定一个')
  full_path = os.path.abspath(filename)
  dir_name = fn(full_path)
  os.makedirs(dir_name, exist_ok=True)
  df = rdf(full_path)
  for i, df_part in enumerate(df_iter(df, chunksize=chunksize, parts=parts)):
    savefile = os.path.join(
        dir_name, f'part_{str(i + 1).zfill(len(str(parts)))}.csv')
    print(f'文件保存至：{savefile}, Rows：{df_part.shape[0]}')
    df_part.to_csv(savefile, index=0)



import os

import pandas as pd

from ..util.decorator import print_doc
from ..util.docx import Docx
from ..util.util import ensure_list
from .extract import rdf
from .stat import describe_auto
from .stat import suspect_series_type


class DataReporter(Docx):
  """数据检测并生成报告"""

  def __init__(
      self, data: (str, pd.DataFrame),
      only: list = None,
      exclude: list = None):
    """
    Args:
      data: 待检测的数据，文件路径或Dataframe
      only: 要检测的列
      exclude: 要排除的列，如果only有值，则以only为准
    """
    self.set_default_style()
    self.doc.add_heading('数据检测报告', 0)
    if isinstance(data, str):
      if os.path.exists(data):
        self.add_intense_quote(f'Data：{data}')
        self.df = rdf(data, info=True)
      else:
        raise FileNotFoundError(f'未找到文件{data}')
    elif isinstance(data, pd.DataFrame):
      self.df = data
    else:
      raise ValueError('请输入Dataframe或路径')

    columns = list(self.df.columns)
    if not only:
      only = columns
    elif exclude:
      only = [c for c in columns if c not in exclude]

    self.only = ensure_list(only)
    self.df = self.df[only]

  @print_doc
  def preprocess(self):
    """数据预处理"""
    for c in self.only:
      if self.df[c].dtype not in (float, int, 'datetime64[ns]'):
        try:
          self.df[c] = pd.to_datetime(self.df[c])
        except Exception:
          pass

  @print_doc
  def basic(self):
    """基础信息描述"""
    self.add_normal_p(f'检测时间：{self.create_time}')
    # 列名
    self.add_title_1('列名')
    self.add_normal_p('，'.join(self.df.columns))
    # 文件size
    self.add_title_1('文件尺寸')
    self.add_normal_p(f'行：{self.df.shape[0]}，列：{self.df.shape[1]}')

    # 列类型
    self.add_title_1('字段名称及类型')
    col_types = pd.DataFrame(
        self.df.dtypes,
        columns=['类型']
    ).reset_index().rename(
        columns={'index': '列名'})
    self.add_table_from_df(col_types)

  @print_doc
  def col_by_col(self):
    """逐列检测"""
    self.add_title_1('字段检测明细')
    for col in self.df.columns:
      self.add_title_3(f'字段：{col}')

      # 空列直接跳过，不进行后续的检查
      if self.df[col].isna().all():
        self.add_paragraph_red('该列为空')
        continue

      self.add_normal_p(f'字段类型：{self.df[col].dtype}')
      self.add_table_from_df(describe_auto(self.df, col), True)

      # 为数值型的列绘制频率分布直方图
      if self.df[col].dtype in (float, int):
        self.add_hist_from_data(
            self.df[~self.df[col].isna()][col].values, True
        )
      # 当列的类型为非数值和日期型时，推断可能的类型
      if self.df[col].dtype not in (float, int, 'datetime64[ns]'):
        if text := suspect_series_type(self.df, col):
          self.add_paragraph_red(text)

  def examine_all(self, file_path):
    """整套流程"""
    self.preprocess()
    self.basic()
    self.col_by_col()
    self.save(file_path=file_path)



import time

from tqdm import tqdm


def to_str(func):
  def wrapper(self):
    if self.dst_format:
      return func(self).strftime(self.dst_format)
    else:
      return func(self)

  return wrapper


def geom_progress(func):
  def wrapper(*args, **kwargs):
    tqdm.pandas(desc=func.__name__.lstrip('geom_'))
    return func(*args, **kwargs)

  return wrapper


def progress(func):
  def wrapper(*args, **kwargs):
    tqdm.pandas(desc=func.__name__)
    return func(*args, **kwargs)

  return wrapper


def print_doc(func):
  def wrapper(*args, **kwargs):
    print(func.__doc__)
    return func(*args, **kwargs)

  return wrapper


def timer(func):
  def wrapper(*args, **kwargs):
    start_time = time.time()
    result = func(*args, **kwargs)
    end_time = time.time()
    print(f'costs：{end_time - start_time:.2f} s')
    return result

  return wrapper
